#ifndef PROGRAMLOADER_TEST_H
#define PROGRAMLOADER_TEST_H

#include <QtTest>

class TestProgramLoader : public QObject {
    Q_OBJECT

public slots:
    void program_loader();
};

#endif // PROGRAMLOADER_TEST_H
